import logging
from dataclasses import dataclass, field
from typing import Any

from livekit.agents import (
    Agent,
    AgentSession,
    JobContext,
    JobProcess,
    RunContext,
    WorkerOptions,
    cli,
    function_tool,
)
from livekit.plugins import (
    elevenlabs,
    groq,
    openai,
    silero,
)

from .db import AsyncDB, CustomerRepository, MaintenanceRepository
from .settings import settings
from .utils import load_prompt

logger = logging.getLogger(__name__)


@dataclass
class UserData:
    personas: dict[str, Any] = field(default_factory=dict)
    prev_agent: Agent | None = None
    ctx: JobContext | None = None
    db: AsyncDB | None = None

    # Customer information
    customer_id: int | None = None
    first_name: str | None = None
    last_name: str | None = None

    def is_identified(self) -> bool:
        return self.first_name is not None and self.last_name is not None

    def is_registered(self) -> bool:
        return self.customer_id is not None

    def reset(self) -> None:
        self.first_name = None
        self.last_name = None
        self.customer_id = None

    def summarize(self) -> str:
        return "User data: Medical office triage system"


RunContext_T = RunContext[UserData]


class BaseAgent(Agent):
    async def on_enter(self) -> None:
        agent_name = self.__class__.__name__
        logger.info(f"Entering {agent_name} context...")

        userdata: UserData = self.session.userdata

        if userdata.ctx and userdata.ctx.room:
            await userdata.ctx.room.local_participant.set_attributes({"agent": agent_name})

        chat_ctx = self.chat_ctx.copy()

        chat_ctx.add_message(
            role="system", content=f"You are the {agent_name}. {userdata.summarize()}"
        )
        await self.update_chat_ctx(chat_ctx)
        self.session.generate_reply()

    def _truncate_chat_ctx(
        self,
        items: list,
        keep_last_n_messages: int = 6,
        keep_system_message: bool = False,
        keep_function_call: bool = False,
    ) -> list:
        def _valid_item(items) -> bool:
            if not keep_system_message and item.type == "message" and item.role == "system":
                return False
            if not keep_function_call and item.type in ["function_call", "function_call_output"]:
                return False
            return True

        new_items = []
        for item in reversed(items):
            if _valid_item(item):
                new_items.append(item)
            if len(new_items) >= keep_last_n_messages:
                break

        while new_items and new_items[0].type in ["function_call", "function_call_output"]:
            new_items.pop(0)

        return new_items

    async def _transfer_to_agent(self, name: str, context: RunContext_T) -> Agent:
        userdata = context.userdata
        current_agent = context.session.current_agent
        next_agent = userdata.personas[name]
        userdata.prev_agent = current_agent

        return next_agent


class TriageAgent(BaseAgent):
    def __init__(
        self,
    ) -> None:
        super().__init__(
            instructions="Greet the user",
            stt=groq.STT(api_key=settings.groq.api_key.get_secret_value()),
            llm=openai.LLM(
                base_url="https://openrouter.ai/api/v1",
                api_key=settings.openrouter.api_key.get_secret_value(),
                model=settings.openrouter.model,
            ),
            tts=elevenlabs.TTS(
                api_key=settings.elevenlabs.api_key.get_secret_value(),
                voice_id=settings.elevenlabs.voice_id,
            ),
            vad=silero.VAD.load(),
        )

    @function_tool
    async def transfer_to_inquiry(self, context: RunContext_T) -> Agent:
        userdata = self.session.userdata
        if userdata.is_identified():
            message = f"Thank you, {userdata.first_name}. Transferring you to our Agent that can help you with further queries"
        else:
            message = "I'll transfer you to our Agent who can help you with your inquiry."

        await self.session.say(message)
        return await self._transfer_to_agent("inquiry", context)


class InquiryAgent(BaseAgent):
    def __init__(self) -> None:
        super().__init__(
            instructions=load_prompt("inquiry_prompt.yaml"),
            stt=groq.STT(api_key=settings.groq.api_key.get_secret_value()),
            llm=openai.LLM(
                model=settings.openrouter.model,
                api_key=settings.openrouter.api_key.get_secret_value(),
                base_url="https://openrouter.ai/api/v1",
            ),
            tts=elevenlabs.TTS(
                voice_id=settings.elevenlabs.voice_id,
                api_key=settings.elevenlabs.api_key.get_secret_value(),
            ),
            vad=silero.VAD.load(),
        )

    @function_tool
    async def identify_customer(self, first_name: str, last_name: str):
        userdata: UserData = self.session.userdata
        userdata.first_name = first_name
        userdata.last_name = last_name

        if not userdata.db:
            return "It seems I'm having trouble connecting with the database"

        async with userdata.db.db_session() as session:
            async with CustomerRepository(session) as repository:
                customer = await repository.get_by_name(first_name=first_name, last_name=last_name)
                if not customer:
                    return (
                        f"I'm sorry, {first_name}, it doesn't seem like your name is in the system"
                    )
                userdata.customer_id = customer.id

        return f"Thank you, {first_name}. I've found your customer data"

    @function_tool
    async def get_maintenance_history(self):
        userdata: UserData = self.session.userdata
        if not userdata.is_identified():
            return "Please identify the current customer using the identify_customer function"

        if not userdata.is_registered():
            return "The current customer does not exist in our records, please relay this to them."

        if not userdata.db:
            return "It seems I'm having trouble connecting with the database"

        async with userdata.db.db_session() as session:
            async with MaintenanceRepository(session) as repository:
                history = await repository.get_maintenance_history(customer_id=userdata.customer_id)

        return history

    @function_tool
    async def transfer_to_triage(self, context: RunContext_T) -> Agent:
        userdata: UserData = self.session.userdata
        if userdata.is_identified():
            message = f"Thank you, {userdata.first_name}. I'll transfer you back to our Triage agent who can better direct your inquiry."
        else:
            message = (
                "I'll transfer you back to our Triage agent who can better direct your inquiry."
            )

        await self.session.say(message)
        return await self._transfer_to_agent("triage", context)


def prewarm(proc: JobProcess):
    proc.userdata["db"] = AsyncDB()


async def entrypoint(ctx: JobContext):
    await ctx.connect()

    userdata = UserData(ctx=ctx)

    triage_agent = TriageAgent()
    inquiry_agent = InquiryAgent()

    userdata.personas.update(
        {
            "triage": triage_agent,
            "inquiry": inquiry_agent,
        }
    )

    session = AgentSession[UserData](userdata=userdata)

    await session.start(
        agent=triage_agent,
        room=ctx.room,
    )

    async def db_dispose():
        if hasattr(ctx.proc.userdata, "db") and ctx.proc.userdata["db"] is not None:
            await ctx.proc.userdata["db"].close()

    ctx.add_shutdown_callback(db_dispose)


if __name__ == "__main__":
    cli.run_app(
        WorkerOptions(
            prewarm_fnc=prewarm,
            entrypoint_fnc=entrypoint,
            ws_url=settings.livekit.url,
            api_key=settings.livekit.api_key.get_secret_value(),
            api_secret=settings.livekit.api_secret.get_secret_value(),
        )
    )
