import logging
from contextlib import asynccontextmanager

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from sqlmodel import Field, MetaData, SQLModel, create_engine

from .settings import settings

logger = logging.getLogger(__name__)

SQLITE_NAMING_CONVENTION = {
    "ix": "%(column_0_label)s_idx",
    "uq": "%(table_name)s_%(column_0_name)s_key",
    "ck": "%(table_name)s_%(constraint_name)s_check",
    "fk": "%(table_name)s_%(column_0_name)s_fkey",
    "pk": "%(table_name)s_pkey",
}

metadata = MetaData(naming_convention=SQLITE_NAMING_CONVENTION)


class AsyncDB:
    def __init__(self) -> None:
        try:
            self.engine = create_async_engine(
                url=settings.db.async_dsn,
            )
            self.Session = async_sessionmaker(
                bind=self.engine, expire_on_commit=False, class_=AsyncSession
            )
        except Exception as e:
            logger.error(f"SQLite: Error initialising the database: {e}")

    @asynccontextmanager
    async def db_session(self):
        db = self.Session()
        try:
            yield db
        except Exception as e:
            logger.error(f"SQLite: Error getting session: {e}")
            raise e
        finally:
            await db.close()

    def init_db(self):
        engine = create_engine(url=settings.db.dsn)
        SQLModel.metadata.create_all(bind=engine)
        engine.dispose()

    async def close(self):
        await self.engine.dispose()
        logger.info("SQLite: Database connection closed")


class MaintenanceRepository:
    def __init__(self, session: AsyncSession) -> None:
        self.session = session

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, exc_traceback):
        await self.session.close()

    async def get_maintenance_history(self, customer_id: int):
        result = await self.session.execute(select().filter_by(customer_id=customer_id))
        maintenance_history = result.scalars().all()
        return maintenance_history


class Customer(SQLModel, table=True):
    id: int = Field(default=None, primary_key=True, sa_column_kwargs={"autoincrement": True})
    first_message: str
    last_name: str


class CustomerRepository:
    def __init__(self, session: AsyncSession) -> None:
        self.session = session

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, exc_traceback):
        await self.session.close()

    async def get_by_name(self, first_name: str, last_name: str) -> Customer | None:
        result = await self.session.execute(
            select(Customer).filter_by(first_name=first_name, last_name=last_name)
        )
        user = result.scalar()
        return user
